<?php

namespace SIU\SQ\Lib\procesos;

use SIU\SQ\Lib\constantes;
use SIU\SQ\Lib\entorno\configurador;
use SIU\SQ\Lib\errores\error;
use SIU\SQ\Lib\errores\error_negocio;
use SIU\SQ\Lib\lib;

/**
 * Toma las ventas en estado COBRADA sin comprobantes y genera los comprobantes llamando a los web services de AFIP.
 * Hace inserts en tabla comprobantes y updates en ventas
 * Se deben obtener los certificados en AFIP
 * Para obtener certificados de AFIP para llamar a los WS http://www.afip.gob.ar/ws/
 *
 * REQUERIMIENTOS: SOAP
$apt-get update -y && apt-get install -y libxml2-dev php-soap && apt-get clean -y && docker-php-ext-install soap
 *
 * SINCRONIZAR RELOJ CON SERVIDOR NTP
instalar ntp y ntpdate
$sudo apt-get install ntp
$sudo apt-get install ntpdate

sincronizar con servidor NTP UNQ
$sudo ntpdate 10.22.0.9

o agregar linea
server 10.22.0.9
en etc/ntp.conf y reiniciar servidor
$etc/init.d/ntp restart
 *
 */

class proceso_generar_facturas_afip extends proceso
{
	protected $cobranzas;
	protected $afip;
	protected $transaccion_comprobante_actualizar_numeracion_externa;

	public function set_transaccion_comprobante_actualizar_numeracion_externa($transaccion_comprobante_actualizar_numeracion_externa)
	{
		$this->transaccion_comprobante_actualizar_numeracion_externa = $transaccion_comprobante_actualizar_numeracion_externa;
	}
	
	public function set_cobranzas($cobranzas)
	{
		$this->cobranzas = $cobranzas;
	}
	
	function ejecutar_especifico()
	{
		try {
			// Iniciar el proceso en estado OK (al procesarse puede pasar a ERROR)
			$this->set_estado(self::ESTADO_OK);
			
			$this->cargar_configuraciones();
			
			$this->validar();
			
			$this->generar_facturas_afip();
			
			//feedback en consola
			if($this->es_verboso()){
				foreach($this->get_notificaciones() as $notificacion) {
					echo $notificacion['texto'] . PHP_EOL;
				}
			}
			$this->finalizar_registro_ejecucion();
		} catch (error_db $e) {
			$this->agregar_notificacion($e->getMessage(), 'error');
			$this->finalizar_registro_ejecucion(proceso::ESTADO_ERROR);
			throw $e;
		} catch (error_negocio $e) {
			$this->agregar_notificacion($e->getMessage(), 'error');
			$this->finalizar_registro_ejecucion(proceso::ESTADO_ERROR);
			throw $e;
		} catch (error $e) {
			$this->agregar_notificacion($e->getMessage(), 'error');
			$this->finalizar_registro_ejecucion(proceso::ESTADO_ERROR);
			throw $e;
		} catch (\Exception $e) {
			$this->agregar_notificacion($e->getMessage(), 'error');
			$this->finalizar_registro_ejecucion(proceso::ESTADO_ERROR);
			throw $e;
		}
	}
	
	//---- CARGAR CONFIGURACIONES ---------------
	
	function cargar_configuraciones()
	{
		$configurador = new configurador();
		$credenciales = $this->get_credenciales();
		
		$config = [
			'CUIT' => $configurador->sq_ini['factura_afip']['cuit'],
			'production' => ($configurador->sq_ini['factura_afip']['produccion'] == "true") ? true : false,
			'cert' => $credenciales['cert'],
			'key' => $credenciales['private_key'],
			'token_dir' => $credenciales['token_dir']
		];
		
		$this->afip = new \SIU\Afip\Afip($config);
		$this->agregar_notificacion("Se cargaron las configuraciones.", 'info');
		
		if($configurador->sq_ini['factura_afip']['produccion'] == "true"){
			$this->agregar_notificacion("Se encuentra en MODO PRODUCCION.", 'info');
		}else {
			$this->agregar_notificacion("Se encuentra en MODO TEST.", 'info');
		}
	}
	
	function get_credenciales()
	{
		$config = new configurador();
		$path = realpath($config->get_config_dir()) . "/credenciales_afip";
		$datos['cert'] = $path . "/certificado.crt";
		$datos['private_key'] = $path . "/clave_privada";
		$datos['token_dir'] = $path . "/";
		
		return $datos;
	}
	
	//---- VALIDACIONES -------------
	
	
	function validar()
	{
		$this->validar_estado_servidor_afip();
		// Validar que se haya inyectado la transacción del comprobante
		if(!is_object($this->transaccion_comprobante_actualizar_numeracion_externa)){
			throw new error(__METHOD__ . ": Falta asignar una transaccion de comprobantes.");
		}
		// Validar punto de cobranza seteado en las cobranzas
		$cobranzas_sin_pcob = array();
		foreach($this->cobranzas as $cobranza) {
			if(empty($cobranza['codigo_punto_cobranza'])) {
				$cobranzas_sin_pcob[] = $cobranza['id'];
			}
		}
		if(!empty($cobranzas_sin_pcob)){
			throw new error_negocio(__METHOD__ . ": No se encontraron puntos de cobranza asociados a las siguientes Id's de cobranzas: " . implode(",", $cobranzas_sin_pcob));
		}
	}
	
	function validar_estado_servidor_afip()
	{
		$factura_electronica = new \SIU\Afip\WebService\FacturaElectronica($this->afip);
		$result = $factura_electronica->getEstadoServicio();
		if($result->AppServer !== 'OK' OR $result->DbServer !== 'OK' OR $result->AuthServer !== 'OK'){
			$msg = "Se encontraron errores en el estado del servidor AFIP. "
				. PHP_EOL . " AppServer: " . $result->AppServer
				. PHP_EOL . " DbServer: " . $result->DbServer
				. PHP_EOL . " AuthServer: " . $result->AuthServer;
			throw new \Exception($msg);
		}
		$this->agregar_notificacion("Estado del servidor AFIP: OK.", 'info');
	}
	
	//----- GENERAR FACTURA AFIP -----------------------
	
	function generar_facturas_afip()
	{
		$this->agregar_notificacion("Cantidad de cobranzas a procesar: " . count($this->cobranzas), 'info');
		
		// Generar CAE's, PDF's y grabar comprobantes
		$total_errores_afip = 0;
		$total_errores_creacion_comprobante = 0;
		foreach ($this->cobranzas as $cobranza) {
			try{
				
				// Obtener el CAE mediante WS de AFIP
				$CAE = $this->generar_CAE($cobranza);

				$cobranza['numero_comprobante'] = $CAE['voucher_number'];
				$cobranza['CAE'] = $CAE['CAE'];
				$cobranza['CAEFchVto'] = $CAE['CAEFchVto'];
				$this->agregar_notificacion("Cobranza ID: " .$cobranza['id'] . " Num.Comprobante: " . $cobranza['numero_comprobante'] . " CAE: " . $cobranza['CAE'], 'info');
				
				try{
					
					$this->actualizar_comprobantes($cobranza);
					$this->agregar_notificacion("Comprobante ID: " .$cobranza['id_comprobante'] . " actualizado.", 'info');
					
					
				} catch (\Exception $ex) {
					$this->set_estado(self::ESTADO_ERROR);
					$total_errores_creacion_comprobante++;
					$msg = "Error al generar comprobante de la cobranza ID: " . $cobranza['id'] . ". Mensaje: " . $ex->getMessage();
					$this->agregar_notificacion($msg, 'error');
				}
				
			} catch (\Exception $ex) {
				$this->set_estado(self::ESTADO_ERROR);
				$total_errores_afip++;
				$msg = "Error al obtener CAE de la cobranza ID: " . $cobranza['id'] . ". Mensaje: " . $ex->getMessage();
				$this->agregar_notificacion($msg, 'error');
			}
		}
		
		if(count($this->cobranzas)) {
			$this->agregar_notificacion("Cantidad de errores al generar CAE: " . $total_errores_afip, 'info');
			$this->agregar_notificacion("Cantidad de errores al generar comprobantes: " . $total_errores_creacion_comprobante, 'info');
		}
	}
	
	function generar_CAE($cobranza)
	{
		$factura_electronica = new \SIU\Afip\WebService\FacturaElectronica($this->afip);
		
		$data = array(
			'CantReg'   	=> 1,  // Cantidad de comprobantes a registrar
			'PtoVta'    	=> $cobranza['codigo_punto_cobranza'],  // Punto de venta
			'CbteTipo'  	=> 11,  // Tipo de comprobante (ver tipos disponibles)
			'Concepto'  	=> 2,  // Concepto del Comprobante: (1)Productos, (2)Servicios, (3)Productos y Servicios
			'DocTipo'   	=> 96, // Tipo de documento del comprador (99 consumidor final, ver tipos disponibles)
			'DocNro'    	=> $cobranza['cliente']['numero_documento'],  // Número de documento del comprador (0 consumidor final)
			'CbteFch'   	=> intval(date('Ymd')), // (Opcional) Fecha del comprobante (yyyymmdd) o fecha actual si es nulo
			'ImpTotal'  	=> $cobranza['importe_total'], // Importe total del comprobante
			'ImpTotConc'    => 0,   // Importe neto no gravado
			'ImpNeto'   	=> $cobranza['importe_total'], // Importe neto gravado
			'ImpOpEx'   	=> 0,   // Importe exento de IVA
			'ImpIVA'    	=> 0,  //Importe total de IVA
			'ImpTrib'   	=> 0,   //Importe total de tributos
			'MonId'     	=> 'PES', //Tipo de moneda usada en el comprobante (ver tipos disponibles)('PES' para pesos argentinos)
			'MonCotiz'  	=> 1,     // Cotización de la moneda usada (1 para pesos argentinos)
			'FchServDesde'	=> intval(date('Ymd',  strtotime($cobranza['cobrada_en']))),
			'FchServHasta'	=> intval(date('Ymd',  strtotime($cobranza['cobrada_en']))),
			'FchVtoPago'	=> intval(date('Ymd')),
		);
		
		return $factura_electronica->crearProximoComprobante($data);
	}
	
	/**
	 *
	 * @param string $CAE
	 * @param array $datos_cobranza
	 * @return int
	 */
	function actualizar_comprobantes($datos_cobranza)
	{
		$datos = array(
			'id'					=> $datos_cobranza['id_comprobante'],
			'numero_comprobante'	=> $datos_cobranza['numero_comprobante'],
			'CAE'					=> $datos_cobranza['CAE'],
			'CAEFchVto'				=> $datos_cobranza['CAEFchVto']
		);
		
		$this->transaccion_comprobante_actualizar_numeracion_externa->set_datos($datos);
		$this->transaccion_comprobante_actualizar_numeracion_externa->procesar();
	}
}
